#ifndef INC_sqlite3TokenTypes_hpp_
#define INC_sqlite3TokenTypes_hpp_

/* $ANTLR 2.7.7 (20160127): "sqlite3.g" -> "sqlite3TokenTypes.hpp"$ */

#ifndef CUSTOM_API
# define CUSTOM_API
#endif

#ifdef __cplusplus
struct CUSTOM_API sqlite3TokenTypes {
#endif
	enum {
		EOF_ = 1,
		ABORT = 4,
		ACTION = 5,
		AUTOINCREMENT = 6,
		AS = 7,
		ASC = 8,
		AND = 9,
		OR = 10,
		BETWEEN = 11,
		CASCADE = 12,
		CASE_T = 13,
		CAST = 14,
		CHECK = 15,
		CREATE = 16,
		COLLATE = 17,
		CONFLICT = 18,
		CONSTRAINT = 19,
		CURRENT_TIME = 20,
		CURRENT_DATE = 21,
		CURRENT_TIMESTAMP = 22,
		DEFAULT = 23,
		DEFERRABLE = 24,
		DEFERRED = 25,
		DELETE = 26,
		DESC = 27,
		ELSE_T = 28,
		END = 29,
		ESCAPE = 30,
		FAIL = 31,
		FOREIGN = 32,
		GLOB = 33,
		KEY = 34,
		LIKE = 35,
		TABLE = 36,
		IF_T = 37,
		IGNORE = 38,
		IN = 39,
		INDEX = 40,
		INITIALLY = 41,
		INSERT = 42,
		IMMEDIATE = 43,
		IS = 44,
		NO = 45,
		NOT = 46,
		NULL_T = 47,
		MATCH = 48,
		EXISTS = 49,
		ON = 50,
		PRIMARY = 51,
		RAISE = 52,
		REFERENCES = 53,
		REGEXP = 54,
		REPLACE = 55,
		RESTRICT = 56,
		ROLLBACK = 57,
		ROWID = 58,
		SET = 59,
		TEMPORARY = 60,
		TEMP = 61,
		THEN = 62,
		UNIQUE = 63,
		UPDATE = 64,
		USING = 65,
		VIRTUAL = 66,
		WHEN = 67,
		WHERE = 68,
		WITHOUT = 69,
		TYPE_NAME = 70,
		COLUMNDEF = 71,
		COLUMNCONSTRAINT = 72,
		TABLECONSTRAINT = 73,
		CREATETABLE = 74,
		CREATEINDEX = 75,
		INDEXEDCOLUMN = 76,
		KEYWORDASTABLENAME = 77,
		KEYWORDASCOLUMNNAME = 78,
		DIGIT = 79,
		DOT = 80,
		ID = 81,
		QUOTEDID = 82,
		QUOTEDLITERAL = 83,
		NUMERIC = 84,
		NL = 85,
		COMMENT = 86,
		WS = 87,
		STRINGLITERAL = 88,
		LPAREN = 89,
		RPAREN = 90,
		COMMA = 91,
		SEMI = 92,
		PLUS = 93,
		MINUS = 94,
		STAR = 95,
		TILDE = 96,
		AMPERSAND = 97,
		BITOR = 98,
		OROP = 99,
		EQUAL = 100,
		EQUAL2 = 101,
		GREATER = 102,
		GREATEREQUAL = 103,
		LOWER = 104,
		LOWEREQUAL = 105,
		UNEQUAL = 106,
		UNEQUAL2 = 107,
		BITWISELEFT = 108,
		BITWISERIGHT = 109,
		SELECT = 110,
		SLASH = 111,
		PERCENT = 112,
		NULL_TREE_LOOKAHEAD = 3
	};
#ifdef __cplusplus
};
#endif
#endif /*INC_sqlite3TokenTypes_hpp_*/
